import os
from util import RemoteSequence, RemoteMap, Current

os_chdir = os.chdir
os_fchdir = os.fchdir


class Vim(object):

    def foreach_rtp(self, cb):
        """
        Call the given callable for each path in 'runtimepath' until either 
        callable returns something but None, the exception is raised or there 
        are no longer paths. If stopped in case callable returned non-None, 
        vim.foreach_rtp function returns the value returned by callable.
        """
        for path in self.list_runtime_paths():
            try:
                if cb(path) != None:
                    break;
            except:
                break

    def chdir(self, dir_path):
        """
        Run os.chdir, then all appropriate vim stuff.
        """
        os_chdir(dir_path)
        self.change_directory(dir_path)

    def fchdir(self, dir_fd):
        """
        Run os.chdir, then all appropriate vim stuff.
        """
        os_fchdir(dir_fd)
        self.change_directory(os.getcwd())

    @property
    def buffers(self):
        if not hasattr(self, '_buffers'):
            self._buffers = RemoteSequence(self,
                                           self.Buffer,
                                           lambda: self.get_buffers())
        return self._buffers

    @property
    def windows(self):
        if not hasattr(self, '_windows'):
            self._windows = RemoteSequence(self,
                                           self.Window,
                                           lambda: self.get_windows())
        return self._windows

    @property
    def tabpages(self):
        if not hasattr(self, '_tabpages'):
            self._tabpages = RemoteSequence(self,
                                            self.Tabpage,
                                            lambda: self.get_tabpages())
        return self._tabpages

    @property
    def current(self):
        if not hasattr(self, '_current'):
            self._current = Current(self)
        return self._current

    @property
    def vars(self):
        if not hasattr(self, '_vars'):
            self._vars = RemoteMap(lambda k: self.get_var(k),
                                   lambda k, v: self.set_var(k, v))
        return self._vars

    @property
    def vvars(self):
        if not hasattr(self, '_vvars'):
            self._vvars = RemoteMap(lambda k: self.get_vvar(k),
                                    None)
        return self._vvars

    @property
    def options(self):
        if not hasattr(self, '_options'):
            self._options = RemoteMap(lambda k: self.get_option(k),
                                      lambda k, v: self.set_option(k, v))
        return self._options



class RemoteSequence(object):
    # TODO Need to add better support for this class on the server
    def __init__(self, vim, remote_klass, handle_array_fn):
        self._wrap_fn = lambda handle: remote_klass(vim, handle)
        self._handle_array_fn = handle_array_fn

    def __len__(self):
        return len(self._handle_array_fn())

    def __getitem__(self, idx):
        if not isinstance(idx, slice):
            return self._wrap_fn(self._handle_array_fn()[idx])
        return map(self._wrap_fn, self._handle_array_fn()[idx.start:idx.stop])


    def __iter__(self):
        handles = self._handle_array_fn()
        for handle in handles:
            yield self._wrap_fn(handle)
    
    def __contains__(self, item):
        return item._handle in self._handle_array_fn()


class RemoteMap(object):
    def __init__(self, get_fn, set_fn):
        self._get_fn = get_fn
        self._set_fn = set_fn

    def __getitem__(self, key):
        return self._get_fn(key)

    def __setitem__(self, key, value):
        if not self._set_fn:
            raise TypeError('This dict is read-only')
        self._set_fn(key, value)

    def __delitem__(self, key):
        if not self._set_fn:
            raise TypeError('This dict is read-only')
        return self._set_fn(key, None)

    def __contains__(self, key):
        try:
            self._get_fn(key)
            return True
        except:
            return False


class Current(object):
    def __init__(self, vim):
        self._vim = vim

    @property
    def line(self):
        return self._vim.get_current_line()

    @line.setter
    def line(self, line):
        self._vim.set_current_line(line)

    @property
    def buffer(self):
        return self._vim.get_current_buffer()

    @buffer.setter
    def buffer(self, buffer):
        self._vim.set_current_buffer(buffer)

    @property
    def window(self):
        return self._vim.get_current_window()

    @window.setter
    def window(self, window):
        self._vim.set_current_window(window)

    @property
    def tabpage(self):
        return self._vim.get_current_tabpage()

    @tabpage.setter
    def tabpage(self, tabpage):
        self._vim.set_current_tabpage(tabpage)


class VimError(Exception):
    pass



import msgpack
from collections import deque
from time import time
from mixins import mixins
from util import VimError


class Remote(object):
    """
    Base class for all remote objects(Buffer, Window...).i
    """
    def __init__(self, vim, handle):
        """
        This is the only initializer remote objects need
        """
        self._vim = vim
        self._handle = handle

    def __eq__(self, other):
        return hasattr(other, '_handle') and self._handle == other._handle


class Client(object):
    """
    Neovim client. It depends on a stream, an object that implements two
    methods:
        - read(): Returns any amount of data as soon as it's available
        - write(chunk): Writes data

    Both methods should be fully blocking.
    """
    def __init__(self, stream):
        self._request_id = 0
        self._stream = stream
        self._unpacker = msgpack.Unpacker()
        self.pending_events = deque()
        self.vim = None

    def pop_events(self, max_index=None):
        if max_index is None:
            rv = list(self.pending_events)
            self.pending_events = deque()
        else:
            rv = []
            i = 0
            while i <= max_index:
                rv.append(self.pending_events.popleft())
                i += 1
        return rv

    def next_message(self, timeout=None):
        """
        Returns the next msgpack object from the input stream. This blocks
        until `timeout` or forever if timeout=None
        """
        while True:
            try:
                return self._unpacker.next()
            except StopIteration:
                chunk = self._stream.read(timeout)
                if not chunk:
                    return
                self._unpacker.feed(chunk)

    def msgpack_rpc_request(self, method_id, params):
        """
        Sends a msgpack-rpc request to Neovim and returns the response
        """
        # increment request id
        request_id = self._request_id + 1
        self._stream.write(msgpack.packb([0, request_id, method_id, params]))
        # Enter a loop feeding the unpacker with data until we parse
        # the response
        message = None
        while not message:
            message = self.next_message()
            if message[0] == 2:
                # event, add to the pending queue
                self.pending_events.append(message[1:])
                message = None
        # update request id
        self._request_id = request_id
        return message

    def next_event(self, timeout=None):
        """
        Returns the next server event
        """
        if len(self.pending_events):
            return self.pending_events.popleft()
        message = self.next_message(timeout)
        # The message type must be 2, which is the msgpack-rpc notification
        # type(http://wiki.msgpack.org/display/MSGPACK/RPC+specification). The
        # only other possible message type the server will send is a
        # response(1), but those must be received from `msgpack_rpc_request`
        assert message[0] == 2
        if message:
            return message[1:]

    def expect(self, event_type, timeout=None, predicate=lambda e: True):
        """
        Blocks until an expected event happens. By default, only the event type
        will be used to determine if a received event is being expected. An
        optional predicate(function that receives the event and returns a
        boolean) can be passed, and in this case it must return true for the
        event to be considered as expected.

        It returns a list containing all events up to and including the
        expected event. If a timeout is given and no expected event happens
        until then, an Exception will be raised. 

        This function is inspired by the 'expect' program, and was designed to
        make it simple to test asynchronous editor responses. For example:

        ```python
        def test_insert_mode():
            vim.current.window.cursor = [1, 0]
            vim.command('normal dditest')
            # block until the first line redraws with the string 'test'
            vim.expect('redraw', timeout=2, lambda e: e['lines'][0] == 'test')
        ```
        """
        for i, event in enumerate(self.pending_events):
            if event[0] == event_type and predicate(event):
                return self.pop_events(i)

        while timeout is None or timeout > 0:
            start = time()
            event = self.next_message(timeout)
            if not event:
                break
            event = event[1:]
            self.pending_events.append(event)
            if event[0] == event_type and predicate(event):
                return self.pop_events()
            timeout -= time() - start

        raise Exception('Timed out')

    def discover_api(self):
        """
        Discovers the remote API using the special method '0'. After this
        the client will have a `vim` attribute containing an object
        that implements an interface similar to the one found in the
        python-vim module(legacy python->vim bridge)
        """
        if self.vim:
            # Only need to do this once
            return
        channel_id, api = self.msgpack_rpc_request(0, [])[3]
        api = msgpack.unpackb(api)
        # The 'Vim' class is the main entry point of the api
        classes = {'vim': type('Vim', (), {})}
        setattr(classes['vim'], 'next_event',
                lambda s, *args, **kwargs: self.next_event(*args, **kwargs))
        setattr(classes['vim'], 'expect',
                lambda s, *args, **kwargs: self.expect(*args, **kwargs))
        # Build classes for manipulating the remote structures, assigning to a
        # dict using lower case names as keys, so we can easily match methods
        # in the API.
        for cls in api['classes']:
            klass = type(cls + 'Base', (Remote,), {})
            # Methods of this class will pass an integer representing the
            # remote object as first argument
            classes[cls.lower()] = klass
        # now build function wrappers
        for function in api['functions']:
            # Split the name on underscores, the first part is the class name,
            # the remaining is the function name
            class_name, method_name = function['name'].split('_', 1)
            generate_wrapper(self,
                             classes[class_name],
                             method_name,
                             function['id'],
                             function['return_type'],
                             function['parameters'])
        # Now apply all available mixins to the generated classes
        for name, mixin in mixins.items():
            classes[name] = type(mixin.__name__, (classes[name], mixin,), {})
        # Create the 'vim object', which is a singleton of the 'Vim' class
        self.vim = classes['vim']()
        self.vim.channel_id = channel_id
        # Add attributes for each other class
        for name, klass in classes.items():
            if name != 'vim':
                setattr(self.vim, klass.__name__, klass)

def generate_wrapper(client, klass, name, fid, return_type, parameters):
    """
    Generate an API call wrapper
    """
    # Build a name->pos map for the parameters 
    parameter_names = {}
    parameter_count = 0
    for param in parameters:
        parameter_names[param[1]] = parameter_count
        parameter_count += 1
    # This is the actual function
    @fname(name)
    def rv(*args, **kwargs):
        if isinstance(args[0], client.vim.__class__):
            # functions of the vim object don't need 'self'
            args = args[1:]
        argv = []
        # fill with positional arguments
        for i, arg in enumerate(args):
            if hasattr(client.vim, parameters[i][0]):
                # If the type is a remote object class, we use it's remote
                # handle instead
                arg = arg._handle
            # Add to the argument vector 
            argv.append(arg)
        result = client.msgpack_rpc_request(fid, argv)
        if result[2]:
            # error
            raise VimError(result[2])
        if hasattr(client.vim, return_type):
            # result should be a handle, wrap in it's specialized class
            return getattr(client.vim, return_type)(client.vim, result[3])
        return result[3]
    setattr(klass, name, rv)


def fname(name):
    """
    Helper for renaming generated functions
    """
    def dec(f):
        f.__name__ = name
        return f
    return dec



import sys, pyuv

class UvStream(object):
    """
    Blocking read/write stream implemented with libuv. Provides the read/write
    interface required by `Client`
    """
    def __init__(self, address=None, port=None):
        self._loop = pyuv.Loop()
        self._error = None
        self._data = None
        self._written = True
        self._connected = False
        self._timed_out = False
        self._timer = pyuv.Timer(self._loop)
        self._timeout_cb = self._on_timeout.__get__(self, UvStream)
        self._read_cb = self._on_read.__get__(self, UvStream)
        self._write_cb = self._on_write.__get__(self, UvStream)
        connect_cb = self._on_connect.__get__(self, UvStream)
        # Select the type of handle
        if port:
            # tcp
            self._stream = pyuv.TCP(self._loop)
            self._stream.connect((address, port), connect_cb)
        elif address:
            # named pipe or unix socket
            self._stream = pyuv.Pipe(self._loop)
            self._stream.connect(address, connect_cb)
        else:
            # stdin/stdout
            self._read_stream = pyuv.Pipe(self._loop) 
            self._read_stream.open(sys.stdin.fileno())
            self._write_stream = pyuv.Pipe(self._loop) 
            self._write_stream.open(sys.stdout.fileno())

    """
    Called when the libuv stream is connected
    """
    def _on_connect(self, stream, error):
        if error:
            self._error = IOError(pyuv.errno.strerror(error))
            return
        self._connected = True
        self._read_stream = self._write_stream = stream

    """
    Called when data is read from the libuv stream
    """
    def _on_read(self, handle, data, error):
        if error:
            self._error = IOError(pyuv.errno.strerror(error))
            return
        if not data:
            self._error = IOError('EOF')
            return
        self._data = data

    """
    Called when a timeout occurs
    """
    def _on_timeout(self, handle):
        self._timed_out = True
        self._loop.stop()

    """
    Called when data is written to the libuv stream
    """
    def _on_write(self, handle, error):
        if error:
            self._error = IOError(pyuv.errno.strerror(error))
            return
        self._written = True
    
    """
    Runs the event loop until a certain condition
    """
    def _run(self, condition=lambda: True, timeout=None):
        if timeout == 0 and not condition():
            return

        if timeout:
            self._timer.start(self._timeout_cb, timeout, 0)

        while not (condition() or self._timed_out):
            if self._error:
                if timeout and not self._timed_out:
                    self._timer.stop()
                # Error occurred, throw it to the caller
                err = self._error
                self._error = None
                raise err
            # Continue processing events
            self._loop.run(pyuv.UV_RUN_ONCE)

        if timeout and not self._timed_out:
            self._timer.stop()
        self._timed_out = False

    """
    Read some data
    """
    def read(self, timeout=None):
        # first ensure the stream is connected
        self._run(lambda: self._connected)
        # start reading
        self._read_stream.start_read(self._read_cb)
        # wait until some data is read
        self._run(lambda: self._data, timeout)
        # stop reading
        self._read_stream.stop_read()
        # return a chunk of data
        rv = self._data
        self._data = None
        return rv

    """
    Write some data
    """
    def write(self, chunk):
        # first ensure the stream is connected
        self._run(lambda: self._connected)
        # queue the chunk for writing
        self._write_stream.write(chunk, self._write_cb)
        # unset the written flag
        self._written = False
        # wait for the flag
        self._run(lambda: self._written)




from util import RemoteMap, RemoteSequence

class Tabpage(object):
    @property
    def windows(self):
        if not hasattr(self, '_windows'):
            self._windows = RemoteSequence(self,
                                           self._vim.Window,
                                           lambda: self.get_windows())
        return self._windows

    @property
    def vars(self):
        if not hasattr(self, '_vars'):
            self._vars = RemoteMap(lambda k: self.get_var(k),
                                   lambda k, v: self.set_var(k, v))
        return self._vars

    @property
    def number(self):
        return self._handle

    @property
    def window(self):
        return self.get_window()

    @property
    def valid(self):
        return self.is_valid()



from client import Client
from uv_stream import UvStream

__all__ = ['connect']

def connect(address, port=None):
    client = Client(UvStream(address, port))
    client.discover_api()
    return client.vim



# -*- coding: utf-8 -*-
import os
from nose.tools import with_setup, eq_ as eq, ok_ as ok
from common import vim, cleanup


@with_setup(setup=cleanup)
def test_receiving_events():
    vim.command('call send_event(%d, "test-event", [1,2,3])' % vim.channel_id)
    events = vim.expect('test-event', lambda e: set(e[1]) == set([1, 2, 3]))
    eq(events, [['test-event', [1, 2, 3]]])
    vim.command('au FileType python call send_event(%d, "py!", bufnr("$"))' %
                vim.channel_id)
    vim.command('set filetype=python')
    events = vim.expect('py!', lambda e: e[1] == vim.current.buffer.number)
    eq(events, [['py!', vim.current.buffer.number]])


